/***********************************************************
 * This software is part of the ProM package * http://www.processmining.org/ * *
 * Copyright (c) 2003-2008 TU/e Eindhoven * and is licensed under the * LGPL
 * License, Version 1.0 * by Eindhoven University of Technology * Department of
 * Information Systems * http://www.processmining.org * *
 ***********************************************************/

package org.processmining.plugins.petrinet.behavioralanalysis;

import java.util.Collection;
import java.util.SortedSet;
import java.util.TreeSet;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.behavioral.AbstractSemanticConnection;
import org.processmining.models.connections.petrinets.behavioral.CoverabilityGraphConnection;
import org.processmining.models.connections.petrinets.behavioral.DeadTransitionsConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.analysis.DeadTransitionsSet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.transitionsystem.CoverabilityGraph;
import org.processmining.models.semantics.Semantics;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.PetrinetSemantics;
import org.processmining.models.semantics.petrinet.impl.PetrinetSemanticsFactory;

/**
 * This class analyze dead transitions in a given net Based on coverability
 * graph written by Murata, Tadao. Petri Nets:Properties, Analysis, and
 * Applications. Proceedings of the IEEE vol. 77, No.4, April 1989. A transition
 * T is dead iff it does not appear as an arc label in a coverability graph's
 * transition
 * 
 * @author arya
 * @email arya.adriansyah@gmail.com
 * @version Nov 8, 2008
 */
@Plugin(name = "Find Dead Transition", returnLabels = { "Dead Transitions" }, returnTypes = { DeadTransitionsSet.class }, parameterLabels = {
		"Net", "Marking", "Coverability Graph", "Semantics" }, userAccessible = true)
public class DeadTransitionAnalyzer {

	// variant with net, marking, and coverability graph as input
	@PluginVariant(variantLabel = "Analyze Dead Transition", requiredParameterLabels = { 0, 1, 2 })
	public DeadTransitionsSet analyzeDeadTransitionPetriNet(PluginContext context, Petrinet net, Marking state,
			CoverabilityGraph graph) throws ConnectionCannotBeObtained {
		// check connection between coverability graph, net, and marking
		Semantics<Marking, Transition> sem = null;
		CoverabilityGraphConnection c = context.getConnectionManager().getFirstConnection(
				CoverabilityGraphConnection.class, context, net, graph, state);
		sem = c.getObjectWithRole(AbstractSemanticConnection.SEMANTICS);
		return analyzeDeadTransitionAssumingConnection(context, net, state, graph, sem);

	}

	// variant with only petri net and marking as input
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "A. Adriansyah", email = "a.adriansyah@tue.nl")
	@PluginVariant(variantLabel = "Analyze Dead Transition", requiredParameterLabels = { 0, 1 })
	public DeadTransitionsSet analyzeDeadTransitionPetriNet(PluginContext context, Petrinet net, Marking state)
			throws ConnectionCannotBeObtained {
		return analyzeDeadTransitionPetriNet(context, net, state, PetrinetSemanticsFactory
				.regularPetrinetSemantics(Petrinet.class));
	}

	// variant with only petri net, marking and semantics
	@PluginVariant(variantLabel = "Analyze Dead Transition", requiredParameterLabels = { 0, 1, 3 })
	public DeadTransitionsSet analyzeDeadTransitionPetriNet(PluginContext context, Petrinet net, Marking state,
			PetrinetSemantics semantics) throws ConnectionCannotBeObtained {
		semantics.initialize(net.getTransitions(), new Marking(state));

		// check if there is already available coverability graph
		CoverabilityGraph graph = null;
		graph = context.tryToFindOrConstructFirstObject(CoverabilityGraph.class, CoverabilityGraphConnection.class,
				CoverabilityGraphConnection.STATEPACE, net, state, semantics);

		return analyzeDeadTransitionAssumingConnection(context, net, state, graph, semantics);
	}

	/**
	 * Main method to analyze transition with an assumption that all provided
	 * parameters are connected
	 * 
	 * @param context
	 *            context of the net
	 * @param net
	 *            net to be analyzed
	 * @param state
	 *            initial state (initial marking)
	 * @param graph
	 *            coverability graph
	 * @param semantics
	 *            semantic
	 * @return DeadTransitionsSet set of dead transitions
	 * @throws Exception
	 */
	private DeadTransitionsSet analyzeDeadTransitionAssumingConnection(PluginContext context, Petrinet net,
			Marking state, CoverabilityGraph graph, Semantics<Marking, Transition> semantics) {
		// for each transition, if it's not included in coverability graph, add it to list of dead transition
		SortedSet<Transition> result = new TreeSet<Transition>();

		// complete path is needed as other Transition class from transitionsystem package is also used in the program
		for (Transition trans : net.getTransitions()) {
			Collection<Object> identifiers = graph.getTransitions();
			if (!identifiers.contains(trans)) {
				result.add(trans);
			}

		}
		DeadTransitionsSet nodeMarking = new DeadTransitionsSet();
		nodeMarking.add(result);

		context.addConnection(new DeadTransitionsConnection(net, state, semantics, nodeMarking));
		context.getFutureResult(0).setLabel("Dead Transition of " + net.getLabel());

		return nodeMarking;
	}

}
