/***********************************************************
 * This software is part of the ProM package * http://www.processmining.org/ * *
 * Copyright (c) 2003-2008 TU/e Eindhoven * and is licensed under the * LGPL
 * License, Version 1.0 * by Eindhoven University of Technology * Department of
 * Information Systems * http://www.processmining.org * *
 ***********************************************************/

package org.processmining.plugins.petrinet.structuralanalysis;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;

import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.connections.ConnectionManager;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.collection.MultiSet;
import org.processmining.models.connections.petrinets.structural.AbstractInvariantMarkingConnection;
import org.processmining.models.connections.petrinets.structural.PlaceInvariantConnection;
import org.processmining.models.connections.petrinets.structural.SComponentConnection;
import org.processmining.models.graphbased.directed.petrinet.InhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.analysis.PlaceInvariantSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.SComponentSet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.petrinet.structuralanalysis.invariants.PlaceInvariantCalculator;

import cern.colt.list.DoubleArrayList;
import cern.colt.list.IntArrayList;
import cern.colt.matrix.DoubleMatrix2D;

/**
 * Class to decompose petri net into its S-components (if there is any) this
 * implementation based on characteristic of place invariants:
 * 
 * "any S-component consists of the support of a minimal support P-invariant and the transitions adjacent to these places"
 * 
 * Taken from : D. Hauschildt, E. Verbeek and W. van der Aalst. WOFLAN: A
 * Petri-net-based Workflow Analyzer. Computing Science Reports 97/12.
 * Eindhoven, August 1997. p24.
 * 
 * @author arya
 * @email arya.adriansyah@gmail.com
 * @version Nov 2, 2008
 */
@Plugin(name = "Identify S-Components", returnLabels = { "S-Components" }, returnTypes = { SComponentSet.class }, parameterLabels = {
		"Net", "Basis of place invariants" }, userAccessible = true)
public class SComponentGenerator {

	// variants with place invariant as input
	@PluginVariant(variantLabel = "Identify S-Components", requiredParameterLabels = { 0, 1 })
	public SComponentSet calculateSComponentsPetriNet(PluginContext context, Petrinet net, PlaceInvariantSet invMarking)
			throws ConnectionCannotBeObtained {
		return calculateSComponentsInternal(context, net, invMarking);
	}

	@PluginVariant(variantLabel = "Identify S-Components", requiredParameterLabels = { 0, 1 })
	public SComponentSet calculateSComponentsInhibitorNet(PluginContext context, InhibitorNet net,
			PlaceInvariantSet invMarking) throws ConnectionCannotBeObtained {
		return calculateSComponentsInternal(context, net, invMarking);
	}

	// variants with no input in form of place invariant
	@PluginVariant(variantLabel = "Identify S-Components", requiredParameterLabels = { 0 })
	public SComponentSet calculateSComponentsPetriNet(PluginContext context, Petrinet net)
			throws ConnectionCannotBeObtained {
		return calculateSComponentsInternalWithInvariantAddition(context, net);
	}

	@PluginVariant(variantLabel = "Identify S-Components", requiredParameterLabels = { 0 })
	public SComponentSet calculateSComponentsInhibitorNet(PluginContext context, InhibitorNet net)
			throws ConnectionCannotBeObtained {
		return calculateSComponentsInternalWithInvariantAddition(context, net);
	}

	/**
	 * This method generate invariants or taking an existing invariants from
	 * context in order to continue calculation of S Component
	 * 
	 * @param context
	 *            context of the net
	 * @param net
	 *            net to be calculated
	 * @return SComponentSet set of S-Component in the net
	 * @throws ExecutionException
	 * @throws InterruptedException
	 * @throws CancellationException
	 * @throws ConnectionCannotBeObtained
	 * @throws Exception
	 */
	private SComponentSet calculateSComponentsInternalWithInvariantAddition(PluginContext context, PetrinetGraph net)
			throws ConnectionCannotBeObtained {
		PlaceInvariantSet invMarking = null;

		// check whether there is a transition invariant before
		invMarking = context.tryToFindOrConstructFirstObject(PlaceInvariantSet.class, PlaceInvariantConnection.class,
				AbstractInvariantMarkingConnection.INVARIANTMARKING, net);

		return calculateSComponentsInternal(context, net, invMarking);
	}

	/**
	 * Main method to calculate S-Components (with context)
	 * 
	 * @param context
	 *            context of the net
	 * @param net
	 *            net to be calculated
	 * @param invMarking
	 *            Place invariant of the corresponding net
	 * @return SComponentSet set of S-Component in the net
	 * @throws Exception
	 */
	private SComponentSet calculateSComponentsInternal(PluginContext context, PetrinetGraph net,
			PlaceInvariantSet invMarking) throws ConnectionCannotBeObtained {
		ConnectionManager cm = context.getConnectionManager();
		// check connectivity between invariant and net
		cm.getFirstConnection(PlaceInvariantConnection.class, context, net, invMarking);

		// calculate S-Component
		SComponentSet nodeMarking = calculateSComponents(net, invMarking);

		// add connection for s-component marking
		context.addConnection(new SComponentConnection(net, nodeMarking));
		context.getFutureResult(0).setLabel("S-Components of " + net.getLabel());

		return nodeMarking;
	}

	/**
	 * Main method to calculate S-Component
	 * 
	 * @param net
	 *            PetrinetGraph with its S-components to be calculate
	 * @param invMarking
	 *            Invariant to be used in calculation. If its null, then place
	 *            invariants will be counted
	 * @return SComponentSet set of S-Component in the net
	 */
	private SComponentSet calculateSComponents(PetrinetGraph net, PlaceInvariantSet invMarking) {
		// initialization
		SComponentSet nodeMarking = new SComponentSet(); // to store final
		// result

		// get place invariants
		if (invMarking == null) {
			// calculate place invariants
			PlaceInvariantCalculator calculator = new PlaceInvariantCalculator();
			invMarking = calculator.calculate(net);
		}

		// references
		DoubleMatrix2D incidenceMatrix = IncidenceMatrixFactory.getIncidenceMatrix(net);
		List<Place> placeList = new ArrayList<Place>(net.getPlaces());
		List<Transition> transitionList = new ArrayList<Transition>(net.getTransitions());

		// for each place invariants filter only the one with 1 or 0 as its
		// member
		invariantLoop: for (MultiSet<Place> set : invMarking) {
			// for one set of invariant, check each element
			Set<Integer> placeIndex = new HashSet<Integer>();
			for (PetrinetNode node : set) {
				if (set.occurrences(node) == 1) {
					placeIndex.add(placeList.indexOf(node));
				} else {
					continue invariantLoop;
				}
			}

			// until here, we have an invariant which only consists of 0 and 1

			// iterate the places, generate array of integer to dice
			// incidenceMatrix
			int[] places = new int[placeIndex.size()];
			Iterator<Integer> it = placeIndex.iterator();
			int counter = 0;
			while (it.hasNext()) {
				places[counter] = it.next();
				counter++;
			}
			// iterate all of transition, generate array of integer to dice
			// incidenceMatrix
			int[] transitions = new int[transitionList.size()];
			for (int i = 0; i < transitionList.size(); i++) {
				transitions[i] = i;
			}

			// dice incidence matrix so that it only include necessary place and
			// transitions
			DoubleMatrix2D tempIncidenceMatrix = incidenceMatrix.viewSelection(places, transitions);

			Set<Integer> transitionIndex = new HashSet<Integer>(); // to store
			// result
			// for
			// transition

			// for each columns on the diced incidence matrix, there can only be
			// 2 nonzero value with the sum of 0
			IntArrayList tempTransition = new IntArrayList();
			DoubleArrayList tempTransValue = new DoubleArrayList();
			for (int i = 0; i < tempIncidenceMatrix.columns(); i++) {
				tempIncidenceMatrix.viewColumn(i).getNonZeros(tempTransition, tempTransValue);

				// as only 1 ingoing and 1 outgoing arc is permitted, number of
				// element should be 2
				// as there can only be 1 ingoing and 1 outgoing, addition
				// should be 0
				if ((tempTransition.size() == 2)
						&& (Double.compare(0.0, tempTransValue.get(0) + tempTransValue.get(1)) == 0)) {
					// add to transitionIndex
					transitionIndex.add(i);
				} else if (tempTransition.size() != 0) {
					// this invariant has invalid transition inclusion. Continue
					// to other invariants
					continue invariantLoop;
				}
			}

			// until here, we have our S-components, add all transition and
			// places as an S-component
			SortedSet<PetrinetNode> result = new TreeSet<PetrinetNode>();
			// add all places
			for (int tempIndex : placeIndex) {
				result.add(placeList.get(tempIndex));
			}
			// add transition corresponds to each place
			for (int tempIndex : transitionIndex) {
				result.add(transitionList.get(tempIndex));
			}
			// add to final set
			nodeMarking.add(result);
		}

		return nodeMarking;
	}
}
